#include "curand_kernel.h"

#define pi 3.14159265358979
#define e 2.718281828

#define CauchyDistribution 1
#define ChiSquareDistributionInteger 2
#define ChiSquareDistributionReal 3
#define ExponentialDistribution 4
#define GammaDistribution 5
#define LaplaceDistribution 6
#define LogNormalDistribution 7
#define NormalDistribution 8
#define ParetoDistribution 9
#define StudentTDistribution 10
#define UniformDistribution 11
#define WeibullDistribution 12

__device__ Real_t generateCauchy(curandState* state, Real_t x0, Real_t gamma);
__device__ void generateCenteredMultiNormal(curandState* state, Real_t* corrMat, mint dim, Real_t* out);
__device__ Real_t generateChiSquare(curandState* state, mint k);
__device__ Real_t generateChiSquare(curandState* state, Real_t k);
__device__ Real_t generateExponential(curandState* state, Real_t lambda);
__device__ Real_t generateGamma(curandState* state, Real_t k, Real_t theta);
__device__ Real_t generateLaplace(curandState* state, Real_t mu, Real_t b);
__device__ Real_t generateLogNormal(curandState* state, Real_t mu, Real_t sigma);
__device__ Real_t generateNormal(curandState* state, Real_t mu, Real_t sigma);
__device__ Real_t generatePareto(curandState* state, Real_t xmin, Real_t alpha);
__device__ Real_t generateStudentT(curandState* state, Real_t nu);
__device__ Real_t generateUniform(curandState* state, Real_t a, Real_t b);
__device__ Real_t generateWeibull(curandState* state, Real_t k, Real_t lambda, Real_t theta);

#define cauchydistribution generateCauchy
#define chisquaredistribution generateChiSquare
#define exponentialdistribution generateExponential
#define gammadistribution generateGamma
#define laplacedistribution generateLaplace
#define lognormaldistribution generateLogNormal
#define normaldistribution generateNormal
#define paretodistribution generatePareto
#define studenttdistribution generateStudentT
#define uniformdistribution generateUniform
#define weibulldistribution generateWeibull


__device__ Real_t generateCauchy(curandState* state, Real_t x0, Real_t gamma) {
	Real_t x = curand_uniform(state);
	return x0 + gamma*tan(static_cast<Real_t>(pi)*(x - static_cast<Real_t>(0.5)));
}

__device__ void generateCenteredMultiNormal(curandState* state, Real_t* corrMat, mint dim, Real_t* out) {
	/*mint ii,jj;
	Real_t temp;
	for (ii=0; ii<dim; ii++) {
		out[ii] = generateNormal(state, static_cast<Real_t>(0), static_cast<Real_t>(1));
	}
	for (ii=dim-1; ii>=0; ii--) {
		temp = static_cast<Real_t>(0);
		for (jj=0; jj<=ii; jj++) {
			temp += corrMat[ii*dim + jj]*out[jj];
			out[ii] = temp;
		}
	}*/
	mint ii,jj;
	Real_t temp;
	for (ii=0; ii<dim; ii++) {
		out[ii] = generateNormal(state, static_cast<Real_t>(0), static_cast<Real_t>(1));
	}
	for (ii=0; ii<dim; ii++) {
		temp = static_cast<Real_t>(0);
		for (jj=ii; jj<dim; jj++) {
			temp += corrMat[ii*dim + jj]*out[jj];
		}
		out[ii] = temp;
	}
}

__device__ Real_t generateChiSquare(curandState* state, mint k) {
	Real_t x, res = static_cast<Real_t>(0);
	for (int ii=0; ii<k; ii++) {
		x = curand_normal(state);
		res += x*x;
	}
	return res;
}

__device__ Real_t generateChiSquare(curandState* state, Real_t k) {
	return generateGamma(state, k/static_cast<Real_t>(2), static_cast<Real_t>(2));
}

__device__ Real_t generateExponential(curandState* state, Real_t lambda) {
	Real_t x = curand_uniform(state);
	return -log(x)/lambda;
}

__device__ Real_t generateGamma(curandState* state, Real_t k, Real_t theta) {
	mint integerk, ii;
	Real_t fractionalk, integerres, fractionalres, res;
	
	integerk = floor(k);
	integerres = 0;
	for (ii=0; ii<integerk; ii++) {
		integerres -= log(generateUniform(state, 0.0, 1.0));
	}
	
	Real_t v0, v1, v2, v3, xi, eta;
	fractionalk = k - integerk;
	if (abs(fractionalk) > 0.001) {
		v0 = 2.718281828/(2.718281828 + fractionalk);
		do {
			v1 = generateUniform(state, 0.0, 1.0);
			v2 = generateUniform(state, 0.0, 1.0);
			v3 = generateUniform(state, 0.0, 1.0);
			if (v1 <= v0) {
				xi = pow(v2, 1.0f/fractionalk);
				eta = v3*pow(xi, fractionalk - 1);
			} else {
				xi = 1 - log(v2);
				eta = v3*exp(-xi);
			}
		} while (eta > exp(-xi)*pow(xi, fractionalk - 1));
		fractionalres = eta;
	} else {
		fractionalres = 0;
	}
	
	res = theta*(integerres + fractionalres);
	
	return res;
}

__device__ Real_t generateLaplace(curandState* state, Real_t mu, Real_t b) {
	Real_t x = curand_uniform(state);
	Real_t res = b*log(static_cast<Real_t>(1) - static_cast<Real_t>(2)*abs(x - static_cast<Real_t>(0.5)));
	if (x < 0.5) {
		res = -res;
	}
	return mu - res;
}

__device__ Real_t generateLogNormal(curandState* state, Real_t mu, Real_t sigma) {
	Real_t x = mu + sigma*curand_normal(state);
	return exp(x);
}

__device__ Real_t generateNormal(curandState* state, Real_t mu, Real_t sigma) {
	return mu + sigma*curand_normal(state);
}

__device__ Real_t generatePareto(curandState* state, Real_t xmin, Real_t alpha) {
	return xmin * generateExponential(state, alpha);
}

__device__ Real_t generateStudentT(curandState* state, Real_t nu) {
	Real_t Z, V;
	Z = generateNormal(state, static_cast<Real_t>(0), static_cast<Real_t>(1));
	V = generateChiSquare(state, nu);
	return Z * rsqrt(V/nu);
}

__device__ Real_t generateUniform(curandState* state, Real_t a, Real_t b) {
	Real_t res = curand_uniform(state);
	return (b-a)*res + a;
}

__device__ Real_t generateWeibull(curandState* state, Real_t k, Real_t lambda, Real_t theta) {
	Real_t x = generateExponential(state, static_cast<Real_t>(1));
	Real_t res = pow(x, static_cast<Real_t>(1)/k);
	res *= lambda;
	res += theta;
	return res;
}


struct BrownianMotion {
	Real_t currentVal, currentTime;

	BrownianMotion(curandState* state) : 
		currentVal(static_cast<Real_t>(0)), currentTime(static_cast<Real_t>(0)) {}

	__device__ Real_t step(curandState* state, Real_t mu, Real_t sigma, Real_t newTime) {
		Real_t dt = newTime - currentTime;
		currentVal += generateNormal(state, dt*mu, sqrt(dt)*sigma);
		currentTime = newTime;
		return currentVal;
	}
};

#if 0
struct JumpDiffusionProcess {
	Real_t currentVal, currentTime, nextJump, lambda;

	JumpDiffusionProcess(curandState* state) : 
		currentVal(static_cast<Real_t>(0)), currentTime(static_cast<Real_t>(0))
	{
		nextJump = generateExponential(state, lambda);
	}

	__device__ Real_t step(curandState* state, Real_t mu, Real_t sigma, Real_t jmu, Real_t jsigma, Real_t newTime) {
		Real_t dt = newTime - currentTime;
		currentVal += generateNormal(state, dt*mu, sqrt(dt)*sigma);
		currentTime = newTime;

		while (currentTime > nextJump) {
			currentVal += generateNormal(state, jmu, jsigma);
			nextJump += generateExponential(state, lambda);
		}

		return currentVal;
	}
};
#endif /* 0 */

extern "C" __global__ void generateSamples(mint distribution, mint seed, mint n, Real_t* out, Real_t x, Real_t y, Real_t z, mint k) {
	mint tid = threadIdx.x + blockDim.x * blockIdx.x;
	mint stride = gridDim.x;
	mint ii;

	curandState rngState;
	curand_init((unsigned long long)seed, tid, 0, &rngState);

	for (ii = tid; ii < n; ii += stride) {
		switch(distribution) {
			case CauchyDistribution:
				out[ii] = generateCauchy(&rngState, x, y);
				break;
			case ChiSquareDistributionInteger:
				out[ii] = generateChiSquare(&rngState, k);
				break;
			case ChiSquareDistributionReal:
				out[ii] = generateChiSquare(&rngState, x);
				break;
			case ExponentialDistribution:
				out[ii] = generateExponential(&rngState, x);
				break;
			case GammaDistribution:
				out[ii] = generateGamma(&rngState, x, y);
				break;
			case LaplaceDistribution:
				out[ii] = generateLaplace(&rngState, x, y);
				break;
			case LogNormalDistribution:
				out[ii] = generateLogNormal(&rngState, x, y);
				break;
			case NormalDistribution:
				out[ii] = generateNormal(&rngState, x, y);
				break;
			case ParetoDistribution:
				out[ii] = generatePareto(&rngState, x, y);
				break;
			case StudentTDistribution:
				out[ii] = generateStudentT(&rngState, x);
				break;
			case UniformDistribution:
				out[ii] = generateUniform(&rngState, x, y);
				break;
			case WeibullDistribution:
				out[ii] = generateWeibull(&rngState, x, y, z);
				break;
			default:
				out[ii] = static_cast<Real_t>(0.0);
				break;
		}
	}
}


